//is a < b
function defaultCompare(a, b) {
  if(a === undefined && b !== undefined)               return true;
  if(a === null && b !== null)                         return true;
  if(a !== undefined && b === undefined)               return false;
  if(a !== null && b === null)                         return false;
  if(typeof a === 'boolean' && typeof b === 'boolean') return a && !b;
  if(typeof a === 'boolean' && typeof b !== 'boolean') return true;
  if(typeof a !== 'boolean' && typeof b === 'boolean') return false;
  if(typeof a === 'number'  && typeof b === 'number')  return a < b;
  if(typeof a === 'string'  && typeof b === 'number')  return a.localeCompare(new Number(b).toString());
  if(typeof a === 'number'  && typeof b === 'string')  return new Number(a).toString().localeCompare(b);
  if(typeof a === 'string'  && typeof b === 'string')  return a.localeCompare(b);
  if(typeof a === 'object')   throw new Error('The default comparator cannot compare objects'); 
  if(typeof a === 'function') throw new Error('The default comparator cannot compare functions'); 
  throw new Error('Oops, missed one: ' + a + ' < ' + b);
}

function Map(root) {
  this.tree = root;
  this.compare = defaultCompare;
}

Map.prototype.insert = function(key, value) {
  if(this.tree === undefined) return new Map(new Tree(this.compare, "b", new Leaf(this.compare), key, value, new Leaf(this.compare)));
  else return new Map(this.tree.insert(key, value));
};

Map.prototype.remove = function(key) {
  if(this.tree !== undefined) {
    var map = new Map(this.tree.remove(key));
    if(map.tree.key === undefined)
      return new Map();
    return map;
  }
  return this;
};

Map.prototype.find = function(key) {
  if(this.tree === undefined) return undefined;
  else return this.tree.find(key);
};

/**
 * color is a string "r", "b", "bb", or "-b"
 */
function Tree(compare, color, left, key, value, right) {
  this.compare = compare;
  this.color   = color;
  this.left    = left;
  this.key     = key;
  this.value   = value;
  this.right   = right;
}

Tree.prototype.find = function(key) {
  var node = this;
  do {
    if(node instanceof Leaf)
      return undefined;

    if(node.key == key)
      return node.value;

    if(this.compare(key, node.key)) node = node.left;
    else node = node.right;
  } while (node);
};

Tree.prototype.insert = function(key, value) {
  var stack = new Array();
  
  var node = this,
      root = null,
      copy = null,
      left = false;

  node = this;
  if(node.compare(key, node.key)) { // <
    if(node.left instanceof Leaf) {
      root = new Tree(this.compare, "b", 
		      new Tree(this.compare, "r", new Leaf(this.compare), key, value, new Leaf(this.compare)), 
		      node.key, node.value, 
		      node.right);

      return root;
    } else {
      root = new Tree(this.compare, node.color,
		      new Tree(this.compare, node.left.color, null, node.left.key, node.left.value, null), 
		      node.key, node.value, 
		      node.right);

      node = node.left;
      copy = root.left;
      stack.push(root);
      stack.push(copy);
    }
  } else if(node.compare(node.key, key)) { // >
    if(node.right instanceof Leaf) {
      root = new Tree(this.compare, "b", 
		      node.left,
		      node.key, node.value, 
		      new Tree(this.compare, "r", new Leaf(this.compare), key, value, new Leaf(this.compare)));
      return root;
    } else {
      root = new Tree(this.compare, node.color, 
		      node.left,
		      node.key, node.value, 
		      new Tree(this.compare, node.right.color, null, node.right.key, node.right.value, null));

      node = node.right;
      copy = root.right;
      stack.push(root);
      stack.push(copy);
    }
  } else { // ==
    return new Tree(this.compare, this.color, this.left, key, value, this.right);
  }


  for(;;) {
    if(node.compare(key, node.key)) { // <
      copy.right = node.right;
      if(node.left instanceof Leaf) {
	copy.left = new Tree(this.compare, "r", new Leaf(this.compare), key, value, new Leaf(this.compare));
	stack.push(copy.left);
	break;
      } else {
	copy.left = new Tree(this.compare, node.left.color, null, node.left.key, node.left.value, null);
	copy = copy.left;
	node = node.left;
	stack.push(copy);
      }
    } else if(node.compare(node.key, key)) { // >
      copy.left = node.left;
      if(node.right instanceof Leaf) {
	copy.right = new Tree(this.compare, "r", new Leaf(this.compare), key, value, new Leaf(this.compare));
	stack.push(copy.right);
	break;
      } else {
	copy.right = new Tree(this.compare, node.right.color, null, node.right.key, node.right.value, null);
	copy = copy.right;
	node = node.right;
	stack.push(copy);
      }
    } else { // ==
      copy.value = value;
      copy.left = node.left;
      copy.right = node.right;
      root.color = "b";
      return root;
    } 
  }

  while(stack.length > 0) {
    balance(stack.pop());
  }
  /**/

  root.color = "b";
  return root;
};

Tree.prototype.remove = function(key) {
  var stack = new Array();
  
  var node = this,
      copy = new Tree(this.compare, node.color, null, node.key, node.value, null),
      root = copy,
      target = null,
      targetNode = null,
      dir = "ACK",
      prev = null;

  stack.push(copy);

  //Find the target node
  for(;;) {
    if(node.compare(key, node.key)) { // <
      copy.right = node.right;
      if(node.left instanceof Leaf) {
	//Key not found
	return this;
      } else {
	copy.left = new Tree(this.compare, node.left.color, null, node.left.key, node.left.value, null);
	dir = "left";
	prev = copy;
	copy = copy.left;
	node = node.left;
	stack.push(copy);
      }
    } else if(node.compare(node.key, key)) { // >
      copy.left = node.left;
      if(node.right instanceof Leaf) {
	//Key not found
	return this;
      } else {
	copy.right = new Tree(this.compare, node.right.color, null, node.right.key, node.right.value, null);
	dir = "right";
	prev = copy;
	copy = copy.right;
	node = node.right;
	stack.push(copy);
      }
    } else { // ==
      //Found the node to delete
      target = copy;
      targetNode = node;
      target.left = targetNode.left;

      break;
    } 
  }

  if(targetNode.right instanceof Leaf && targetNode.left instanceof Leaf) {
    if(prev == null) {
      target.key = undefined; //signal to the containing Map that this node is to be deleted
    } else {
      if(targetNode.color == "r") prev[dir] = new Leaf(this.compare, "b");
      else prev[dir] = new Leaf(this.compare, "bb");
    }
  } else if(targetNode.right instanceof Leaf) {
    if(prev == null) {
      target.key = targetNode.left.key;
      target.value = targetNode.left.value;
      target.left = targetNode.left.left;
      target.right = targetNode.left.right;
      return root;
    } else {
      prev[dir] = targetNode.left;
      if(targetNode.left.color == "b")
	targetNode.left.color = targetNode.color;
    }
  } else if(targetNode.left instanceof Leaf) {
    if(prev == null) {
      target.key = targetNode.right.key;
      target.value = targetNode.right.value;
      target.left = targetNode.right.left;
      target.right = targetNode.right.right;
      return root;
    } else {
      prev[dir] = targetNode.right;
      if(targetNode.right.color == "b")
	targetNode.right.color = targetNode.color;
    }
  } else {
    //Has two subtrees
    dir = "right";
    copy.right = new Tree(this.compare, node.right.color, null, node.right.key, node.right.value, null);
    prev = copy;
    node = node.right;
    copy = copy.right;
    stack.push(copy);

    //Find the leftmost node on the right branch
    for(;;) {
      copy.right = node.right;
      if(node.left instanceof Leaf) {
	//found the leftmost node
	break;
      } else {
	copy.left = new Tree(this.compare, node.left.color, null, node.left.key, node.left.value, null);
	dir = "left";
	prev = copy;
	copy = copy.left;
	node = node.left;
	stack.push(copy);
      }
    }

    target.key = node.key;
    target.value = node.value;
    
    if(node.right instanceof Leaf) {
      if(node.color == "r") prev[dir] = new Leaf(this.compare, "b");
      else prev[dir] = new Leaf(this.compare, "bb");
    } else {
      prev[dir] = node.right;
    }

    if(prev.color == "b")
      prev.color = "r";
  }

  while(stack.length > 0) {
    var n = stack.pop();
    bubble(n);
    balance(n);
  }
  /**/

  return root;
}; //end of remove()

function bubble(tree) {
  if(tree.left && tree.right && (tree.left.color == "bb" || tree.right.color == "bb")) {
    tree.color = plusBlack(tree.color);
    tree.left.color = minusBlack(tree.left.color);
    tree.right.color = minusBlack(tree.right.color);
  }
}

function plusBlack(color) {
  switch(color) {
    case "b": return "bb";
    case "r": return "b";
    case "-b": return "r";
  }
}

function minusBlack(color) {
  switch(color) {
    case "bb": return "b";
    case "b": return "r";
    case "r": return "-b";
  }
}

function balance(tree) {
  if(tree.color === "b" || tree.color === "bb") {
    try {
      /* Target:
       *     y
       *    / \
       *   x   z
       *  / \ / \
       * a  b c  d
       */
      var y, x, z, a, b, c, d;

      if(tree.left.color == "r" && tree.left.right.color == "r") {
	/*    (z)
	 *    / \
	 *  (x)  d
	 *  / \
	 * a  (y)
	 *    / \
	 *   b   c
	 */
	z = tree;
	x = z.left;
	d = z.right;
	a = x.left;
	y = x.right;
	b = y.left;
	c = y.right;

	//swap z and y
	var tmp = {
	  key: z.key,
	  value: z.value
	};

	z.key = y.key;
	z.value = y.value;
	y.key = tmp.key;
	y.value = tmp.value;

	tmp = z;
	z = y;
	y = tmp;

	y.left = x;
	y.right = z;	
x.left = a;
	x.right = b;
	z.left = c;
	z.right = d;

	if(tree.color == "b") y.color = "r";
	else y.color = "b";
	x.color = "b";
	z.color = "b";
      } else if(tree.right.color == "r" && tree.right.left.color == "r") {
	/*  (x)
	 *  / \
	 * a  (z)
	 *    / \
	 *  (y)  d
	 *  / \
	 * b   c
	 */

	x = tree;
	a = x.left;
	z = x.right;
	y = z.left;
	d = z.right;
	b = y.left;
	c = y.right;

	//swap x and y
	tmp = {
	  key: x.key,
	  value: x.value
	};

	x.key = y.key;
	x.value = y.value;
	y.key = tmp.key;
	y.value = tmp.value;

	tmp = x;
	x = y;
	y = tmp;

	y.left = x;
	y.right = z;
	x.left = a;
	x.right = b;
	z.left = c;
	z.right = d;

	if(tree.color == "b") y.color = "r";
	else y.color = "b";
	x.color = "b";
	z.color = "b";
      } else if(tree.left.color == "r" && tree.left.left.color == "r") {
	/*      (z)
	 *      / \
	 *    (y)  d
	 *    / \
	 *  (x)  c
	 *  / \
	 * a   b
	 */	

	z = tree;
	y = z.left;
	d = z.right;
	x = y.left;
	c = y.right;
	a = x.left;
	b = x.right;

	//swap z and y
	tmp = {
	  key: z.key,
	  value: z.value
	};

	z.key = y.key;
	z.value = y.value;
	y.key = tmp.key;
	y.value = tmp.value;

	tmp = z;
	z = y;
	y = tmp;

	y.left = x;
	y.right = z;
	x.left = a;
	x.right = b;
	z.left = c;
	z.right = d;

	if(tree.color == "b") y.color = "r";
	else y.color = "b";
	x.color = "b";
	z.color = "b";
      } else if(tree.right.color == "r" && tree.right.right.color == "r") {
	/*  (x)
	 *  / \
	 * a  (y)
	 *    / \
	 *   b  (z)
	 *      / \
	 *     c   d
	 */	

	x = tree;
	a = x.left;
	y = x.right;
	b = y.left;
	z = y.right;
	c = z.left;
	d = z.right;
	
	//swap x and y
	tmp = {
	  key: x.key,
	  value: x.value
	};

	x.key = y.key;
	x.value = y.value;
	y.key = tmp.key;
	y.value = tmp.value;

	tmp = x;
	x = y;
	y = tmp;

	y.left = x;
	y.right = z;
	x.left = a;
	x.right = b;
	z.left = c;
	z.right = d;

	if(tree.color == "b") y.color = "r";
	else y.color = "b";
	x.color = "b";
	z.color = "b";
      }
    } catch(e) {}
  }
}

//"b" or "bb"
function Leaf(compare, color) {
  if(color === undefined)
    color = "b";

  this.compare = compare;
  this.color = color;
}

Leaf.prototype.find = function() {
  return undefined;
};

Leaf.prototype.insert = function(key, value) {
  return new Tree(this.compare, "r", new Leaf(this.compare), key, value, new Leaf(this.compare));
};

module.exports = Map;
