var assert = require("chai").assert,
    Map = require("../index").Map,
    treeify = require("treeify");

describe("Map", function() {
  it("should be instantiable", function() {
    new Map();
  });

  describe(".find", function() {
    it("should return undefined for non-existent keys", function() {
      assert.equal(new Map().find("foo"), undefined);
    });
  });

  describe(".insert", function() {
    it("should not do bad things", function() {
      var map = new Map();

      for(var i = 0; i < 10000; i++) {
	map = map.insert(i, i);
      }

      for(i = 0; i < 10000; i++) {
	assert.equal(map.find(i), i);
      }
    });

    it("should play nicely with random data", function() {
      var map = new Map();
      var numbers = [];
      
      for(var i = 0; i < 5; i++) {
	var rand = Math.floor(Math.random()*1000-500);
	map = map.insert(rand, rand);
	numbers.push(rand);
      }

      numbers.forEach(function(i) {
	assert.equal(map.find(i), i);
      });
    });

    it("should replace an existing key value pair", function() {
      var map = new Map().insert("foo", "bar");
      assert.equal(map.find("foo"), "bar");
      map = map.insert("foo", "baz");
      assert.equal(map.find("foo"), "baz");
    });
  });

  describe(".remove", function() {
    it("should get rid of a key in the map", function() {
      var numbers = [8, 4, 12, 3, 6, 10, 14, 1, 3, 5, 7, 8, 11, 13, 15];
      var map = new Map();

      numbers.forEach(function(num) {
	map = map.insert(num, num);
      });

      numbers.forEach(function(num) {
	assert.equal(map.find(num), num);
      });

      numbers.forEach(function(num) {
	console.log(num);
	map = map.remove(num);
	console.log(treeify.asTree(map, true, true));
	assert.equal(map.find(num), undefined);
      });
    });
  });
});

